/*
 * Data.java
 *
 */

import java.util.ArrayList;
import java.util.Random;
import java.io.*;
import java.util.StringTokenizer;

public class Data {

  private int size;
  private int imgXSize;
  private int imgYSize;
  private ArrayList<int[]> patterns;
  private Random rand;

  /**
   * The Data class. This constructor does not create any patterns. It
   * takes one argument, the seed to the random number generator. If
   * seed == 0 time is used, otherwise seed is used.
   *
   * @param seed Seed to random number generator.
   */
  public Data(int seed) {
    size = 0;
    imgXSize = 0;
    imgYSize = 0;
    patterns = new ArrayList<int[]>();
    rand = (seed == 0 ? new Random(): new Random(seed));
  }
	
  /**
   * The Data class. This constructor does not create any patterns and
   * the current time is used as seed to the random number generator.
   */
    public Data() { this(0); }

  /**
   * The Data class. This creates nPatt random patterns with size size
   * using seed as the seed to the random number generator.
   * 
   * @param size Size of each pattern.
   * @param nPatt Number of patterns to create.
   * @param seed Seed to random number generator.
   */
    public Data(int size, int nPatt, int seed) {
    this(seed);
    this.size = size;
    createRandomPatterns(size, nPatt);
  }

  /**
   * The Data class. This creates nPatt random patterns with size size
   * using the current time as the seed to the random number
   * generator.
   * 
   * @param size Size of each pattern.
   * @param nPatt Number of patterns to create.
   */
    public Data(int size, int nPatt) {this(size, nPatt, 0);}
        
  /**
   * Create nPatt random {-1,1} patterns of size size. Note this method
   * clears any previously stored patterns.
   * 
   * @param size The size of the patterns.
   * @param nPatt The number of patterns to create.
   */
  public void createRandomPatterns(int size, int nPatt) {
    this.size = size;
    patterns.clear();
    for (int j = 0; j < nPatt; ++j) {
      int[] pattern = new int[size];
      for (int i = 0; i < pattern.length; ++i) {
	  pattern[i] = (rand.nextBoolean() ? 1 : -1);
      }
      patterns.add(pattern);
    }
  }

  /**
   * Add nPatt random patterns.
   * 
   * @param nPatt Number of patterns to add.
   */
  public void addRandomPatterns(int nPatt) {
    for (int j = 0; j < nPatt; ++j) {
      int[] pattern = new int[size];
      for (int i = 0; i < pattern.length; ++i) {
	pattern[i] = (rand.nextBoolean() ? 1 : -1);
      }
      patterns.add(pattern);
    }
  }

  /**
   * Load images stored in the file with the name fname.
   * 
   * @param fname Filename of the images to load.
   */
  public void loadImagePatterns(String fname) {
    patterns.clear();
		
    try {
      BufferedReader in = new BufferedReader(new FileReader(fname));
			
      /**
       * Read the first line
       */
      String line = in.readLine();
      StringTokenizer st = new StringTokenizer(line);
      Integer tmp = new Integer(st.nextToken());
      imgXSize = tmp.intValue();
      tmp = new Integer(st.nextToken());
      imgYSize = tmp.intValue();
      size = imgXSize * imgYSize;
			
      /**
       * Read the rest of the lines
       */
      line = in.readLine();
      while (line != null) {
	st = new StringTokenizer(line);
	int[] pattern = new int[size];
	int n = 0;
	while (st.hasMoreTokens()) {
	  tmp = new Integer(st.nextToken());
	  pattern[n++] = tmp.intValue();
	}
	patterns.add(pattern);
	line = in.readLine();
      }
    } catch (Exception e) {
    }
  }

  /**
   * Return all patterns stored.
   * 
   * @return ArrayList of all patterns.
   */
  public ArrayList<int[]> getPatterns() {
    return patterns;
  }

  /**
   * The size of the stored patterns.
   * 
   * @return Pattern size.
   */
  public int getSize() {
    return size;
  }

  /**
   * Get a specific pattern.
   * 
   * @param i The index of the pattern to return.
   * @return The return pattern.
   */
  public int[] getPattern(int i) {
    return (int[]) patterns.get(i);
  }

  /**
   * Get the size (X * Y) of the loaded images.
   * 
   * @return The X and Y size of the images.
   */
  public int[] getImageSize() {
    return (int[]) new int [] {imgXSize, imgYSize};
  }

}
