/*
 * Network.java
 *
 */

public abstract class Network {

  private int sizeN;
  private boolean serialUpdate;
  int[] neurons;
  double[][] weights;

  /**
   * This constructs a new Hopfield model of size number of nodes. 
   *  
   * @param sizeN The number of nodes in the Hopfield model.
   */
  public Network(int sizeN) {
    this.sizeN = sizeN;
    serialUpdate = true;
    neurons = new int[sizeN];
    weights = new double[sizeN][sizeN];
  }

  /**
   * Initialize the network with a specific pattern.
   * 
   * @param pattern Pattern to initialize the nodes with.
   */
  public void initNetwork(int[] pattern) {
    for (int i = 0; i < pattern.length; i++) {
      neurons[i] = pattern[i];
    }
  }

  /**
   * Sets the update procedure.
   * 
   * @param serialUpdate If true then serial update is used, otherwise 
   * parallel update is used.
   */
  public void setUpdateType(boolean serialUpdate) {
    this.serialUpdate = serialUpdate;
  }
	
  /**
   * Train the Hopfield model using the supplied data class.
   * 
   * @param data Data to use when training the Hopfield model.
   */
  public abstract void train(Data data);
		

  /**
   * Update the Hopfield model for a number of iterations (nUpdates). One iteration
   * means updating all nodes once. To update the nodes until convergence of all
   * nodes, use nUpdates = 0.
   * 
   * @param nUpdates The number of iterations to use when updating the nodes.
   */
  public void update(int nUpdates) {
	
    if (serialUpdate)
      updateSerial(nUpdates);
    else
      updateParallel(nUpdates);
  }

  /**
   * Gets the size (ie the number of nodes) of the Hopfield model.
   * 
   * @return The size of the model
   */
  public int getSize() {
    return sizeN;
  }

  /**
   * Return the current state as a vector of integers
   * 
   * @return The current state of the network
   */
  public int[] getState() {
    return (int[])neurons.clone();
  }
	
  /**
   * Find the number of wrong initial bits given the supplied pattern.
   * 
   * @param pattern The supplied pattern to test for initial bit stability.
   * @return The number of unstable bits of the supplied pattern.
   */
  public int getInitialStability(int[] pattern) {
	
    int diff = 0;
    for (int i = 0; i < pattern.length; ++i) {
	
      double v = 0;
      for (int j = 0; j < pattern.length; ++j) {
	v += pattern[j] * weights[i][j];
      }
      if (v != 0) {
	if( Math.abs(Math.signum(v) - pattern[i]) > 0 ) {
	  diff += 1;
	}
      }
    }
    return diff;
  }

  /**
   * This method prints the current state of the network.
   */
  public void printNetwork() {
	
    System.out.printf("Network state:\n");
    for (int i = 0; i < neurons.length; ++i) {
      System.out.printf("%1d ", neurons[i]);
    }
    System.out.printf("\n");
  }

  /**
   * This method prints the weight matrix.
   */
  public void printWeights() {
	
    System.out.printf("Weight matrix:\n");
    for (int i = 0; i < neurons.length; ++i) {
	for (int j = 0; j < neurons.length; ++j) {
	    System.out.printf("%-10.2e ", weights[i][j]);
	}
	System.out.printf("\n");
    }
  }

  /*
   * Private methods
   */
	
  /**
   * Seriel updating of the network.
   * 
   * @param nUpdates The number of iterations to perform. If nUpdates=0
   * then the updating stops when convergence is reached.
   */
  private void updateSerial(int nUpdates) {

    if( nUpdates == 0 )
      updateSerialUntilConvergence();
    else
      updateSerialFixedNumber(nUpdates);
		
  }
	
  /**
   * Serial update until convergence.
   */
  private void updateSerialUntilConvergence() {
		
    MyUtil util = new MyUtil();
    int ok = 0;
    int numUpdates = 0;
    while (ok == 0) {
					
      int[] rnd = util.shuffle(neurons.length);
      int[] old = neurons.clone();
      for (int i = 0; i < neurons.length; ++i) {
	int ii = rnd[i];
	double v = 0;
	for (int j = 0; j < neurons.length; ++j) {
	  v += neurons[j] * weights[ii][j];
	}
	if (v != 0) {
	  neurons[ii] = (int) Math.round(v / Math.abs(v));
	}
      }
      numUpdates++;
      int diff = 0;
      for (int i = 0; i < neurons.length; i++) {
	if( neurons[i] != old[i] ) diff += 1;
      }
      if( diff == 0 ) ok = 1;
    }
    numUpdates--;
    //System.out.printf("Number of updates: %d", numUpdates);
  }

  /**
   * Serial updates for a fixed number of iterations.
   * 
   * @param nUpdates Number of iterations to perform.
   */
  private void updateSerialFixedNumber(int nUpdates) {
		
    for (int n = 0; n < nUpdates; n++) {

      for (int i = 0; i < neurons.length; ++i) {

	double v = 0;
	for (int j = 0; j < neurons.length; ++j) {
	  v += neurons[j] * weights[i][j];
	}
	if (v != 0) {
	  neurons[i] = (int) Math.round(v / Math.abs(v));
	}
      }
    }
  }

  /**
   * Parallel updating of the network for a fixed number of iterations.
   * 
   * @param nUpdates The number of parallel iterations.
   */
  private void updateParallel(int nUpdates) {

    for (int n = 0; n < nUpdates; n++) {

      double[] v = new double[neurons.length];

      for (int i = 0; i < neurons.length; ++i) {
	v[i] = 0;
	for (int j = 0; j < neurons.length; ++j) {
	  v[i] += neurons[j] * weights[i][j];
	}
      }

      for (int i = 0; i < neurons.length; ++i) {
	if (v[i] != 0) {
	  neurons[i] = (int) Math.round(v[i] / Math.abs(v[i]));
	}
      }
    }
		
  }
	
}
