/*
 * PlotPanel.java
 *
 * Created on March 3, 2008, 5:24 PM
 *
 */

import javax.swing.JPanel;
import java.util.*; //BS added
import java.awt.*;
import java.awt.font.*;

/**
 * The PlotPanel plots a set of Plotable as curves with matching axes and tics. 
 * The user can specify what colors the different curves should have. Labels for 
 * the axes can also be specified. 
 * 
 * The PlotPanel should be wrapped in another panel whose size will determine the 
 * size of the PlotPanel.
 * 
 * @author simon
 */
public class PlotPanel extends JPanel
{
  //TODO: - Allow user-specified ranges and tics for x and y-axes
  //      - Allow plotting data as points (scatter plot)
    
  private double xMax,  xMin,  yMax,  yMin;
  private double xFactor,  yFactor;
  private int BORDER_X = 60,  BORDER_Y = 40,  TIC_LENGTH = 10,  PANEL_HEIGHT,  PANEL_WIDTH;
  private int originX,  originY;
  private java.util.LinkedHashMap<Plotable, Color> curves;
  private java.util.Vector<Point> points;
  private String xLabel = "",  yLabel = "";

  /**
   * Creates a new instance of PlotPanel
   */
  public PlotPanel()
  {
    curves = new java.util.LinkedHashMap<Plotable, Color>();
    points = new java.util.Vector<Point>();
  }

  //BS added:
  /**
   * Returns the curves with colors
   */
  public LinkedHashMap<Plotable, Color> getCurves() {
    return curves;
  }

  //BS added:
  /**
   * Returns the curve count
   */
  public int getNumCurves() {
    return curves.size();
  }

  //BS added (for mouse-induced drawing):
  /**
   * Converts pixel coords to user coords
   * @param i the horisontal pixel position
   * @param j the vertical pixel position
   */
  public double[] toXY(int i, int j) {
    double[] x = new double[2];
    x[0] = xMin + (i - BORDER_X) / xFactor;
    x[1] = yMin + (BORDER_Y + PANEL_HEIGHT - j) / yFactor;
    return x;
  }

  /** 
   * adds a new set of data to be included in the plot using the 
   * foreground color of the current panel
   * @param p the Plotable
   */
  public void addPlotable(Plotable p)
  {
    addPlotable(p, getForeground());
  }

  /**
   * adds a new set data to included in the plot, using a custom color
   * @param p the Plotable
   * @param c the color
   */
  public void addPlotable(Plotable p, Color c)
  {
    curves.put(p, c);
    if (curves.size() == 1) {
      xMax = p.getXMax();
      xMin = p.getXMin();
      yMax = p.getYMax();
      yMin = p.getYMin();
    }
  }
    
  /**
   * adds a new point, drawn as a circle, to the plot at a given position
   * @param x x-coordinate
   * @param y y-coordiante
   * @param radius radius of point, given in pixels
   * @param c color of the point
   * @param filled true if the point is to be filled
   */
    
  public void addPoint(double x, double y, int radius, Color c, 
		       boolean filled)
  {
    Point p = new Point();
    p.x = x;
    p.y = y;
    p.radius = radius;
    p.color = c;
    p.filled = filled;
    points.add(p);
  }
  /**
   * Set a label for the x-axis
   * @param s the label
   */
  public void setXLabel(String s)
  {
    xLabel = s;
  }

  /**
   * Set a label for the y-axis
   * @param s the label
   */
  public void setYLabel(String s)
  {
    yLabel = s;
  }

  /**
   * Remove all data from the plot
   */
  public void reset()
  {
    curves.clear();
    points.clear();
  }
    
  public void clear() 
  {
    java.awt.Color fg = getForeground();
    setForeground(getBackground());
    getGraphics().fillRect(0, 0, getWidth(), getHeight());
    setForeground(fg);
  }
    
  /**
   * {@inheritDoc}
   */
  @Override
    public void paintComponent(Graphics g)
  {
    setSize(getParent().getSize());
    plot(g);
  }

  /** 
   * Plot data to Graphics object g
   */
  public void plot(Graphics g)
  {
    PANEL_HEIGHT = getSize().height - 2 * BORDER_Y;
    PANEL_WIDTH = getSize().width - 2 * BORDER_X;
    //initialize graphics
    Graphics2D g2d = (Graphics2D) g;

    g2d.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
					     RenderingHints.VALUE_ANTIALIAS_ON));
    g2d.setColor(getForeground());
    g2d.setBackground(getBackground());
    g2d.clearRect(0, 0, PANEL_WIDTH + 2 * BORDER_X,
		  PANEL_HEIGHT + 2 * BORDER_Y);
    checkRange();
    if( curves.size() > 0 ) { // avoid empty ranges
      drawAxes(g2d);
      drawCurves(g2d);
      drawPoints(g2d);
    }
  }

  /** 
   * Transform x-value to pixel-coordinate
   */
  private int xToPixel(double x)
  {
    return (int) ((x - xMin) * xFactor) + BORDER_X;
  }
    
  /** 
   * Transform y-value to pixel-coordinate
   */
  private int yToPixel(double y)
  {
    return -(int) ((y - yMin) * yFactor) + BORDER_Y + PANEL_HEIGHT;
  }
    
  /**
   * Calculate x- and y-range
   */
  private void checkRange()
  {
    if (curves.size() > 0)
      for (Plotable p : curves.keySet()) {
	if (p.getXMax() > xMax) xMax = p.getXMax();
	if (p.getXMin() < xMin) xMin = p.getXMin();
	if (p.getYMax() > yMax) yMax = p.getYMax();
	if (p.getYMin() < yMin) yMin = p.getYMin();
      }
  }

  /**
   * Draw x- and y-axes, including tics
   */
  private void drawAxes(Graphics2D g2d)
  {
    double yrange = yMax - yMin;
    double xrange = xMax - xMin;

    // Determine ranges and tic distances
    if ( yrange <= 0 ) {
      System.err.println("Error: empty yrange for plot");
      return;
    }
    if ( xrange <= 0 ) {
      System.err.println("Error: empty xrange for plot");
      return;
    }
	
    int orderOfMagnitude;

    /**
     * y-tics
     */
    orderOfMagnitude = (int) Math.floor(Math.log10(yrange));
    double dy = Math.pow(10, orderOfMagnitude);
    //get a decent number of tics...
    if (yrange / dy < 6) dy /= 2;
    if (yrange / dy > 11) dy *= 2;
    //get a bottom value that can be expressed as i*10^n, where i = -10...10;
    double yval = yMin = dy * Math.floor(yMin / dy);
    //find the largest tic
    while (yval < yMax) yval += dy;
    yMax = yval;
    yrange = yMax - yMin;
    yFactor = PANEL_HEIGHT / yrange;

    /**
     * x-tics
     */
    orderOfMagnitude = (int) Math.floor(Math.log10(xrange));
    double dx = Math.pow(10, orderOfMagnitude);
    //get a decent number of tics...
    if (xrange / dx < 4) dx /= 2;
    if (xrange / dx > 10) dx *= 2;
    double xval = xMin = dx * Math.floor(xMin / dx);
    while (xval < xMax) xval += dx;
    xMax = xval;
    xrange = xMax - xMin;
    xFactor = PANEL_WIDTH / xrange;

    //find where to put the axes, they will cross as close to the (0,0) as possible
    if (xMin >= 0) originX = BORDER_X;
    else if (xMax < 0) originX = PANEL_WIDTH + BORDER_X;
    else originX = -(int) (xMin * xFactor) + BORDER_X;

    if (yMin >= 0) originY = BORDER_Y + PANEL_HEIGHT;
    else if (yMax <= 0) originY = BORDER_Y;
    else originY = (int) (yMin * yFactor) + BORDER_Y + PANEL_HEIGHT;

    //paint the axes
    BasicStroke normalStroke = new BasicStroke(1.0f, BasicStroke.CAP_SQUARE,
					       BasicStroke.JOIN_MITER);
    g2d.setColor(getForeground());
    g2d.setStroke(normalStroke);
    //x-axis
    g2d.drawLine(BORDER_X, originY, BORDER_X + PANEL_WIDTH, originY);
    //y-axis
    g2d.drawLine(originX, BORDER_Y, originX, BORDER_Y + PANEL_HEIGHT);

    //x-tics
    if (xMin < 0 && xMax > 0) {
      for (xval = 0; xval >= xMin; xval -= dx) drawXTic(g2d, xval);
      for (xval = dx; xval <= xMax; xval += dx) drawXTic(g2d, xval);
    } else for (xval = xMin; xval <= xMax; xval += dx) drawXTic(g2d, xval);

    //y-tics
    if (yMin < 0 && yMax > 0) {
      for (yval = 0; yval >= yMin; yval -= dy) drawYTic(g2d, yval);
      for (yval = dy; yval <= yMax; yval += dy) drawYTic(g2d, yval);
    } else for (yval = yMin; yval <= yMax; yval += dy) drawYTic(g2d, yval);

    g2d.drawString(xLabel, BORDER_X + (int) (PANEL_WIDTH / 2) - 10,
		   (int) (1.9 * BORDER_Y) + PANEL_HEIGHT);
    if (!yLabel.isEmpty()) {
      java.text.AttributedString rotatedYLabel =
	new java.text.AttributedString(yLabel);
      rotatedYLabel.addAttribute(TextAttribute.TRANSFORM,
				 new TransformAttribute(java.awt.geom.AffineTransform.getRotateInstance(java.lang.Math.PI*1.5)));
      g2d.drawString(rotatedYLabel.getIterator(),
		     (int) (0.2 * BORDER_X),
		     BORDER_Y + (int) (PANEL_HEIGHT / 2) + 20);
    }
  }

  private void drawXTic(Graphics2D g2d, double xval)
  {
    int xtic = xToPixel(xval);
    String tic1 = String.format("%.3g", xval);
    String tic2 = xval + "";
    String tic = tic1.length() > tic2.length() ? tic2 : tic1;
    g2d.drawLine(xtic, originY - (int) (TIC_LENGTH / 2),
		 xtic, originY + (int) (TIC_LENGTH / 2));
    g2d.drawString(tic + "", xtic - 8, originY + 2 * TIC_LENGTH);

  }

  private void drawYTic(Graphics2D g2d, double yval)
  {
    int ytic = yToPixel(yval);
    String tic1 = String.format("%.3g", yval);
    String tic2 = yval + "";
    String tic = tic1.length() > tic2.length() ? tic2 : tic1;
    g2d.drawLine(originX - (int) (TIC_LENGTH / 2), ytic,
		 originX + (int) (TIC_LENGTH / 2), ytic);
    g2d.drawString(tic + "", originX - 4 * TIC_LENGTH, ytic + 5);
  }

  /** 
   * Plot data
   */
  private void drawCurves(Graphics2D g2d)
  {
    BasicStroke normalStroke = new BasicStroke(1.0f, BasicStroke.CAP_SQUARE,
					       BasicStroke.JOIN_MITER);
    g2d.setStroke(normalStroke);
    for (java.util.Map.Entry<Plotable, Color> e : curves.entrySet()) {
      g2d.setColor(e.getValue());
      Plotable p = e.getKey();
      double[] x = p.getX();
      double[] y = p.getY();
      if (x.length != y.length) {
	System.err.println("Size mismatch in Plotable.drawCurves()");
	return;
      }
      for (int i = 0; i < x.length - 1; ++i) {
	int x1 = xToPixel(x[i]);
	int x2 = xToPixel(x[i+1]);
	int y1 = yToPixel(y[i]);
	int y2 = yToPixel(y[i+1]);
	g2d.drawLine(x1, y1, x2, y2);
      }
    }
  } 

  /**
   * Plot points 
   * 
   */
  private void drawPoints(Graphics2D g2d) 
  {
    for (Point p : points) {
      int x = xToPixel(p.x) - p.radius;
      int y = yToPixel(p.y) - p.radius;
      g2d.setColor(p.color);
      if (p.filled) {
	g2d.fillOval(x, y, p.radius*2, p.radius*2);
      }
      else {
	g2d.drawOval(x, y, p.radius*2, p.radius*2);		
      }
    }
  }
    
  class Point 
  {
    double x, y;
    Color color;   
    int radius;
    boolean filled;
  }
}


